/*
 * FILE: clip.c     
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this program does 3D clipping 
*/

#include <stdio.h> 
#include <math.h>
#include <malloc.h>
#include <GL/aux.h> 
#include "view4d.h"

#define BD_X_POS_ONE	0
#define BD_X_NEG_ONE	1
#define BD_Y_POS_ONE	2
#define BD_Y_NEG_ONE	3
#define BD_Z_NEG_ONE    4
#define BD_Z_POS_ONE	5
#define BD_W_ZERO	6

extern void clip3D();
static void SH_clip();
static GLint inside();
static void output();
static GLint intersect();
static void scale_vertex();

extern void
clip3D()
{
	GLint iface, iv;

#ifdef DEBUG_CLIP
        fopen(DEBUGFILE, "a");
#endif

	for (iface = 0; iface < polytope->num_faces; iface++) {
		polytope->face[iface].visible = GL_TRUE;

#ifdef DEBUG_CLIP
		fprintf(dfd," FACE %d:\n", iface);
	        for (iv = 0; iv < polytope->face[iface].num_vertices; iv++) {
		   fprintf(dfd,"  face vertex %d = %d", iv, polytope->face[iface].vertex[iv]);
		   print_point("", polytope->screen_vtx[polytope->face[iface].vertex[iv]]);
	        }
#endif

		SH_clip(iface, BD_W_ZERO);

#ifdef DEBUG_CLIP
	        fprintf(dfd," FACE %d W CLIPPED:\n", iface);
	        for (iv = 0; iv < polytope->face[iface].num_vertices; iv++) {
		   fprintf(dfd,"   face vertex %d = %d", iv, polytope->face[iface].vertex[iv]);
		   print_point("", polytope->screen_vtx[polytope->face[iface].vertex[iv]]);
	        }
#endif

	        SH_clip(iface, BD_X_POS_ONE);
		SH_clip(iface, BD_Y_POS_ONE);
		SH_clip(iface, BD_X_NEG_ONE);
		SH_clip(iface, BD_Y_NEG_ONE);
		SH_clip(iface, BD_Z_POS_ONE); 
		SH_clip(iface, BD_Z_NEG_ONE);
	
#ifdef DEBUG_CLIP
	        fprintf(dfd," FACE %d CLIPPED:\n", iface);
	        for (iv = 0; iv < polytope->face[iface].num_vertices; iv++) {
		   fprintf(dfd,"   face vertex %d = %d", iv, polytope->face[iface].vertex[iv]);
		   print_point("", polytope->screen_vtx[polytope->face[iface].vertex[iv]]);
	        }
                for (iv = 0; iv < polytope->face[iface].num_edges; iv++) {
		   fprintf(dfd," face edge %d = %d: %d %d \n", iv, 
			polytope->face[iface].edge[iv],
			polytope->edge[polytope->face[iface].edge[iv]].a,
			polytope->edge[polytope->face[iface].edge[iv]].b);
		}
#endif
	}

#ifdef DEBUG_CLIP
        fclose(dfd);
#endif
	
}

/********************************************************************/
/* FUNCTION: SH_clip()
   PURPOSE: this function clips the unscaled screen coordinates to the 
	    cube -1<=z,y,x<=1, using the Sutherland Hodgman algorithm
	    listed on p 128 of Foley and van Dam.
*/
/***********************************************************************/
static void SH_clip(face, clip_boundary)
	GLint face, clip_boundary;
{
	GLint j, outlength = 0, inlength, s, p, new_vertex, clipped;
	GLint out_array[MAX_FACE_VERTICES], ic, id, last_vertex;
	struct edge_struct edge;
	

	inlength = polytope->face[face].num_vertices;
	s = polytope->face[face].vertex[inlength - 1];
	clipped = GL_FALSE;

	for (j = 0;j < inlength; j++) {
		
	    p = polytope->face[face].vertex[j];

	    if (inside(p, clip_boundary)) {    /* cases 1 and 4 */
		if (inside(s, clip_boundary)) {         /* case 1 */
			output(p, &outlength, out_array);
		}
		else {                                   /* case 4 */
			new_vertex = intersect(s, p, clip_boundary);
		        output(new_vertex, &outlength, out_array);
			output(p, &outlength, out_array);
			clipped = GL_TRUE;
		}
	    }
	    else {
		if (inside(s, clip_boundary)) {     /* cases 2 and 3 */
		    new_vertex = intersect(s, p, clip_boundary);  /* case 2 */
		    output(new_vertex, &outlength, out_array);   
		    clipped = GL_TRUE; 
		}
		    /* no action for case 3 */
	    }
	    s = p;
	}

	    polytope->face[face].num_vertices = outlength;
	/* eliminate duplicate vertices from out_array (if any) */
	if (clipped) {
	    last_vertex = out_array[outlength - 1];
	    for (ic = 0; ic < polytope->face[face].num_vertices; ic++) {
		if ((FP_NEQ(polytope->screen_vtx[out_array[ic]].p[X_COORD],
		    polytope->screen_vtx[last_vertex].p[X_COORD])) ||
		   (FP_NEQ(polytope->screen_vtx[out_array[ic]].p[Y_COORD],
		    polytope->screen_vtx[last_vertex].p[Y_COORD])) ||
                   (FP_NEQ(polytope->screen_vtx[out_array[ic]].p[Z_COORD],
		    polytope->screen_vtx[last_vertex].p[Z_COORD]))) 
		    polytope->face[face].vertex[ic] = out_array[ic];
		else polytope->face[face].num_vertices--;
		last_vertex = out_array[ic];
	    }

	    /* need to generate new edges */
	    for (ic = 0;ic < (polytope->face[face].num_vertices - 1); ic++) {
		edge.a = polytope->face[face].vertex[ic];
		edge.b = polytope->face[face].vertex[ic+1];
		polytope->edge[polytope->num_edges] = edge;
		polytope->face[face].edge[ic] = polytope->num_edges;
		polytope->num_edges++;
	    }
		/* do wrap around edge */
	    edge.a = polytope->face[face].vertex[
		polytope->face[face].num_vertices - 1];
	    edge.b = polytope->face[face].vertex[0];
	    polytope->edge[polytope->num_edges] = edge;
	    polytope->face[face].edge[ic] = polytope->num_edges;
	    polytope->num_edges++;

	   polytope->face[face].num_edges = polytope->face[face].num_vertices;
	 
	}
	    if (outlength) polytope->face[face].visible = GL_TRUE;
	    else polytope->face[face].visible = GL_FALSE;
		
}
	
static GLint
inside(index, clip_boundary)
	GLint index, clip_boundary;
{
	struct point vertex;


	vertex = polytope->screen_vtx[index];

	switch (clip_boundary)
	{

		case BD_X_POS_ONE:
			if (FP_GT(vertex.p[X_COORD], 1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_X_NEG_ONE:
			if (FP_LT(vertex.p[X_COORD], -1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_Y_POS_ONE:
			if (FP_GT(vertex.p[Y_COORD], 1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_Y_NEG_ONE:
			if (FP_LT(vertex.p[Y_COORD], -1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_Z_NEG_ONE:
			if (FP_LT(vertex.p[Z_COORD], -1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_Z_POS_ONE:
			if (FP_GT(vertex.p[Z_COORD], 1.0)) return GL_FALSE;
			else return GL_TRUE;
		case BD_W_ZERO:
			if (FP_GT(vertex.p[W_COORD], 0.0)) return GL_TRUE;
			else return GL_FALSE;
		default:
			printf("ERROR: bad clip_boundary to inside!!\n");
			exit(1);

	}

}

static void
output(new_vertex, length, out_array)
	int new_vertex, *length, *out_array;
{
	
	out_array[*length] = new_vertex;
	(*length)++;

}

static GLint
intersect(index1, index2, clip_boundary)
	GLint index1, index2, clip_boundary;
{
	struct point vertex1, vertex2;
	GLint new_vertex, temp;
	GLdouble range, scale_fac;

	new_vertex = polytope->num_vertices;
	polytope->num_vertices++;

	polytope->clipped = GL_TRUE;

	switch (clip_boundary)
	{

		case BD_X_POS_ONE:
		    if (!inside(index1, BD_X_POS_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
		    }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[X_COORD] - vertex1.p[X_COORD]);
		    scale_fac = (1.0 - vertex1.p[X_COORD])/range;
		    polytope->screen_vtx[new_vertex].p[X_COORD] = 1.0;
		    scale_vertex(new_vertex, index1, index2, 
				X_COORD, scale_fac);
		    break;
	        case BD_X_NEG_ONE:
	            if (!inside(index1, BD_X_NEG_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
		    }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[X_COORD] - vertex1.p[X_COORD]);
		    scale_fac = (vertex1.p[X_COORD] + 1.0)/range;
		    polytope->screen_vtx[new_vertex].p[X_COORD] = -1.0;
		    scale_vertex(new_vertex, index1, index2, 
				X_COORD, scale_fac);
		    break;
		case BD_Y_POS_ONE:
		   if (!inside(index1, BD_Y_POS_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
	            }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[Y_COORD] - vertex1.p[Y_COORD]);
		    scale_fac = (1.0 - vertex1.p[Y_COORD])/range;
		    polytope->screen_vtx[new_vertex].p[Y_COORD] = 1.0;
		    scale_vertex(new_vertex, index1, index2, 
				Y_COORD, scale_fac);
	            break;
	        case BD_Y_NEG_ONE:
		    if (!inside(index1, BD_Y_NEG_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
		    }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[Y_COORD] - vertex1.p[Y_COORD]);
		    scale_fac = (vertex1.p[Y_COORD] + 1.0)/range;
		    polytope->screen_vtx[new_vertex].p[Y_COORD] = -1.0;
		   scale_vertex(new_vertex, index1, index2, 
				Y_COORD, scale_fac);
		    break;
		case BD_Z_NEG_ONE:
		   if (!inside(index1, BD_Z_NEG_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
	    	    }
		   vertex1 = polytope->screen_vtx[index1];
	           vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[Z_COORD] - vertex1.p[Z_COORD]);
		    scale_fac = (vertex1.p[Z_COORD] + 1.0)/range;
		    polytope->screen_vtx[new_vertex].p[Z_COORD] = -1.0;
		    scale_vertex(new_vertex, index1, index2, 
				Z_COORD, scale_fac);
		    break;			
		case BD_Z_POS_ONE:
		   if (!inside(index1, BD_Z_POS_ONE)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
	            }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[Z_COORD] - vertex1.p[Z_COORD]);
		    scale_fac = (1.0 - vertex1.p[Z_COORD])/range;
		    polytope->screen_vtx[new_vertex].p[Z_COORD] = 1.0;
		    scale_vertex(new_vertex, index1, index2, 
				Z_COORD, scale_fac);
		    break;
		case BD_W_ZERO:
		   if (!inside(index1, BD_W_ZERO)) {
			temp = index1;
			index1 = index2;
			index2 = temp;
	            }
		    vertex1 = polytope->screen_vtx[index1];
	            vertex2 = polytope->screen_vtx[index2];
                    range = fabs(vertex2.p[W_COORD] - vertex1.p[W_COORD]);
		    scale_fac = vertex1.p[W_COORD]/range;
		    polytope->screen_vtx[new_vertex].p[W_COORD] = 0.0;
		    scale_vertex(new_vertex, index1, index2, 
				W_COORD, scale_fac);
		    break;
		default:
			printf("ERROR: bad clip_boundary to intersect!!\n");
			exit(1);

	}
	return(new_vertex);

}		

static void
scale_vertex(new_vertex, index1, index2, coord, scale_fac)
	GLint new_vertex, coord, index1, index2;
	GLdouble scale_fac;
{
	GLint ic;
	struct point vertex1, vertex2;

	vertex1 = polytope->screen_vtx[index1];
	vertex2 = polytope->screen_vtx[index2];

        for (ic = 0; ic < 3; ic++ ) {
          if (ic != coord) 
	     polytope->screen_vtx[new_vertex].p[ic] =
		vertex1.p[ic] + scale_fac*(vertex2.p[ic]- vertex1.p[ic]);		}

	/* now scale 4D stuff along with the screen coords */
	for (ic = 0; ic < 4; ic++ ) {

            polytope->gouraud_norm[new_vertex].p[ic] =
		polytope->gouraud_norm[index1].p[ic] + scale_fac *
		(polytope->gouraud_norm[index2].p[ic] - 
                    polytope->gouraud_norm[index1].p[ic]);
	   normalize(polytope->gouraud_norm[new_vertex].p);

	   polytope->rot_vertex[new_vertex].p[ic] =
		polytope->rot_vertex[index1].p[ic] + scale_fac *
		(polytope->rot_vertex[index2].p[ic] - 
		 polytope->rot_vertex[index1].p[ic]);	
	}

}    
  	    
  
	   
