/*
 * FILE: light.c
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this file contains all the lighting and specular reflection calcs 
 */

#include <stdio.h> 
#include <math.h>
#include <GL/aux.h>
#include "view4d.h"

#define AMB_LGT_INT_RED		125
#define AMB_LGT_INT_GRN		125
#define AMB_LGT_INT_BLU		125
#define INT_PHONG_RED		130
#define INT_PHONG_GRN		130
#define INT_PHONG_BLU		130
#define LIGHT_BRIGHTNESS	1.0
#define CLOSE_TRANSP_FACTOR	close_tsp
#define FAR_AWAY_TRANSP_FACTOR	far_tsp
#define CLOSE_AMB_COEFF 	close_amb
#define FAR_AWAY_AMB_COEFF	far_amb
#define CLOSE_SPEC_COEFF	close_spec
#define FAR_AWAY_SPEC_COEFF	far_spec
#define CLOSE_DIFFUSE_COEFF	close_diffuse
#define FAR_AWAY_DIFFUSE_COEFF	far_diffuse
#define SCALE_FRONT		1.0
#define SCALE_BACK 		0.0
#define Z_FRONT_FRACTION	1.0
#define Z_BACK_FRACTION		0.0
#define TRANSP_POLY_COLOR_IND	0
#define XSPARENCY(x,y,z)  ((z)*(x) + (1.0 - (z))*(y))


/********** external variable definitions ****************/
GLdouble O_DIFFUSE[MAX_FACES][3];
GLdouble O_SPECULAR[MAX_FACES][3];

GLint Lights[][3] = {
	{255, 255, 255},   /* x-axis */
	{255, 255, 255},   /* y-axis */
	{255, 255, 255}, 
	{255, 255, 255},   /* x-axis */
	{255, 255, 255},   /* y-axis */
	{255, 255, 255},   /* z-axis */
	{255, 255, 255},   /* z-axis */
	{255, 255, 255} ,
  	{255, 255, 255}    /* w-axis */
};
GLdouble z_back, z_front, depth_range, close_diffuse, far_diffuse;
GLdouble close_tsp, far_tsp, close_amb, far_amb, close_spec, far_spec;
/*********** function declarations *********************/
static void pxl_color();
static void specular_lighting();
static GLint convert_to_rgb();
extern void compute_x_deltas();
extern void normalize();
extern GLdouble dot();
extern void fill_pixels();
extern void depth_q();

extern void
fill_pixels(Y, x1, x2)
        GLint Y;
        GLdouble x1, x2;
/**********************************************************/
/* FUNCTION: fill_pixels()                                */
/* PURPOSE:  this function colors the pixels on scnline   */
/*    "Y" between x1<=X<x2 with the color of polygon's    */
/*   object                                               */
/**********************************************************/
{
        GLint i, X1, X2;
        GLfloat pixel_color[3], parray[MAX_PIXELS][3];

        X1 = (GLint) floor(x1);
        X2 = (GLint) floor(x2);

        for (i = X1;i < X2;i++) {
	
#ifdef DEBUG_PIXEL
	fprintf(dfd, "pixel at %d is:", i);
#endif

	    if ((i >= 0) && (i <= width_of_window))  
	       pxl_color(pixel_color, i);
	    else {
		pixel_color[0] = 0.;
		pixel_color[1] = 0.;
		pixel_color[2] = 0.;
	    }
	    parray[i][0] = pixel_color[0];
	    parray[i][1] = pixel_color[1];
	    parray[i][2] = pixel_color[2];
	    if (archive_it) 
		image[i][Y] = convert_to_rgb(pixel_color[0], 
			pixel_color[1], pixel_color[2]);

        }
	
	/* draw it */

	glRasterPos2i(X1, Y);
	glDrawPixels(X2 - X1 + 1, 1, GL_RGB, GL_FLOAT, parray[X1]);

}

static GLint
convert_to_rgb(r, g, b)
 GLfloat r, g, b;
{
  GLint hexcolor, ir, ig, ib;

  ir = r * 255;
  ig = g * 255;
  ib = b * 255;

  hexcolor = (ib << 16) + (ig << 8) + ir;
  return(hexcolor);
}

/**********************************************************/
/* FUNCTION: pixel_color()				  */
/* PURPOSE: this function computes the intensity to the   */
/*      current point contributed by ambient lighting     */
/**********************************************************/
static void
pxl_color(dintensity, x)
	GLfloat *dintensity; 
	GLint x;
{
	GLint jface, iface;
        GLdouble x_delta, fraction, pixel_depth, scale_factor, poly_color[3],
		amb_coeff, transp_fac, diffuse_coeff, spec_coeff;

	dintensity[RED_IND] = 190.;
	dintensity[GRN_IND] = 190.;
	dintensity[BLU_IND] = 190.;
	
	/* now draw polygons from front to back */
	for (jface = 0; jface < polytope->num_faces;jface++) {
		iface = polytope->ordered_faces[jface];
		if (PT[iface].IN) {

			x_delta = (double) (x - PT[iface].x_start);
			if (FP_LT(x_delta, 0.0)) x_delta = 0.0;
			pixel_depth = PT[iface].dist_start + 
				x_delta*PT[iface].dist_delta;	
  			if (FP_GT(pixel_depth, z_back)) {
				printf(
				"BAD pixel_depth: %lf x_delta: %lf iface: %d\n", 
					pixel_depth, x_delta, iface);
	 			scale_factor = SCALE_BACK;
        		}
			else if (FP_LT(pixel_depth, z_front)) {
	 			printf( 
				"BAD pixel_dept: %lf x_delta %lf iface: %d\n",
					pixel_depth,x_delta,  iface);
			scale_factor = SCALE_FRONT;
        		}
			else scale_factor = SCALE_BACK + 
				(z_back - pixel_depth)*
                		(SCALE_FRONT - SCALE_BACK)/depth_range;
        		transp_fac = CLOSE_TRANSP_FACTOR*scale_factor + 
				(1. - scale_factor)*FAR_AWAY_TRANSP_FACTOR;
        		amb_coeff =  CLOSE_AMB_COEFF*scale_factor + 
				(1. - scale_factor)*FAR_AWAY_AMB_COEFF;
			diffuse_coeff =  CLOSE_DIFFUSE_COEFF*scale_factor + 
				(1. - scale_factor)*FAR_AWAY_DIFFUSE_COEFF;
			spec_coeff =  CLOSE_SPEC_COEFF*scale_factor + 
				(1. - scale_factor)*FAR_AWAY_SPEC_COEFF;
		
		        /** compute ambient lighting **/
			poly_color[RED_IND] =  AMB_LGT_INT_RED * amb_coeff *
				 O_DIFFUSE[iface][RED_IND];	
			poly_color[GRN_IND] = AMB_LGT_INT_GRN * amb_coeff * 
				O_DIFFUSE[iface][GRN_IND];	
			poly_color[BLU_IND] =  AMB_LGT_INT_BLU * amb_coeff *
				 O_DIFFUSE[iface][BLU_IND];


			/*** compute specular lighting ***/
			specular_lighting(iface, poly_color, x_delta,
				 spec_coeff, diffuse_coeff);

			/* add this polygon's color to pixel's total color */
			dintensity[RED_IND] = XSPARENCY(dintensity[RED_IND],
			   poly_color[RED_IND], transp_fac);
			dintensity[GRN_IND] = XSPARENCY(dintensity[GRN_IND],
			   poly_color[GRN_IND], transp_fac);
			dintensity[BLU_IND] = XSPARENCY(dintensity[BLU_IND],
			   poly_color[BLU_IND], transp_fac);

		}
	}


	if (FP_GT(dintensity[RED_IND], 255.)) dintensity[RED_IND] = 255.;
	if (FP_GT(dintensity[GRN_IND], 255.)) dintensity[GRN_IND] = 255.;
	if (FP_GT(dintensity[BLU_IND], 255.)) dintensity[BLU_IND] = 255.;

	dintensity[RED_IND] = dintensity[RED_IND]/255.;
	dintensity[GRN_IND] = dintensity[GRN_IND]/255.;
	dintensity[BLU_IND] = dintensity[BLU_IND]/255.;
			

#ifdef DEBUG_PIXEL
	fprintf(dfd, "intensity: r: %f g: %f b: %f\n", dintensity[0],
		dintensity[1], dintensity[2]);
#endif

}

/**********************************************************/
/* FUNCTION: specular_lighting()			  */
/* PURPOSE: this function computes the intensity to the   */
/*      current point contributed by specular reflection  */
/**********************************************************/
static void
specular_lighting(poly, poly_color, x_delta, spec_coeff, diffuse_coeff)
	GLint poly;
	GLdouble *poly_color, x_delta, spec_coeff, diffuse_coeff;
{
	GLint ic, il;
	GLdouble cos_alpha, cos_theta, f_att, cos_alpha_sq, cos_theta_sq,
		norm_len_sq, eye_len_sq;
	struct point N, R, V, L, lights[MAX_LIGHTS], rflxs[MAX_LIGHTS];
	
	/* interpolate N, R, L at point x */
	for (ic = 0; ic < 4; ic++) {
	  N.p[ic] = PT[poly].N_start.p[ic] + 
			(PT[poly].N_delta.p[ic]* x_delta);
	  R.p[ic] = PT[poly].R_start.p[ic] + 
			(PT[poly].R_delta.p[ic]* x_delta);
	  V.p[ic] = PT[poly].V_start.p[ic] + 
			(PT[poly].V_delta.p[ic]* x_delta);
	  L.p[ic] = PT[poly].L_start.p[ic] + 
			(PT[poly].L_delta.p[ic]* x_delta);
	  for (il =0; il < NUM_LIGHTS; il++) {
	     lights[il].p[ic] = PT[poly].lghts_start[il].p[ic] +
		(PT[poly].lghts_delta[il].p[ic] * x_delta);
	     rflxs[il].p[ic] = PT[poly].refls_start[il].p[ic] +
		(PT[poly].refls_delta[il].p[ic] * x_delta);
	  }
	}
  	
	norm_len_sq = N.p[X_COORD]*N.p[X_COORD] + N.p[Y_COORD]*N.p[Y_COORD]
		+ N.p[Z_COORD]*N.p[Z_COORD] + N.p[W_COORD]*N.p[W_COORD];
	eye_len_sq = V.p[X_COORD]*V.p[X_COORD] + V.p[Y_COORD]*V.p[Y_COORD]
		+ V.p[Z_COORD]*V.p[Z_COORD] + V.p[W_COORD]*V.p[W_COORD];

 	/* compute phong intensity */
	cos_theta = dot(N.p, L.p);
	cos_alpha = dot(R.p, V.p);

#ifdef DEBUG_PIXEL
	fprintf(dfd, " theta: %lf alpha: %lf diffuse: %lf spec: %lf\n", 
			cos_theta, cos_alpha, diffuse_coeff, spec_coeff);
#endif

	if (FP_LT(cos_alpha, 0.0)) cos_alpha_sq  = 0.0;
	else {
		cos_alpha_sq = (cos_alpha * cos_alpha)/ 
		((R.p[X_COORD]*R.p[X_COORD] + R.p[Y_COORD]*R.p[Y_COORD]
		+ R.p[Z_COORD]*R.p[Z_COORD] + R.p[W_COORD]*R.p[W_COORD]) *
		eye_len_sq);
	}
	if (FP_LT(cos_theta, 0.0)) cos_theta_sq = 0.0; 
	else {
		cos_theta_sq = (cos_theta * cos_theta) / 
		(norm_len_sq *
		(L.p[X_COORD]*L.p[X_COORD] + L.p[Y_COORD]*L.p[Y_COORD]
	       + L.p[Z_COORD]*L.p[Z_COORD] + L.p[W_COORD]*L.p[W_COORD]));
	}

	poly_color[RED_IND] = poly_color[RED_IND] + (INT_PHONG_RED *
		((diffuse_coeff * O_DIFFUSE[poly][RED_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][RED_IND] * cos_alpha_sq)));
	poly_color[GRN_IND] = poly_color[GRN_IND] + (INT_PHONG_GRN *
		((diffuse_coeff * O_DIFFUSE[poly][GRN_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][GRN_IND] * cos_alpha_sq)));
	poly_color[BLU_IND] = poly_color[BLU_IND] + (INT_PHONG_BLU *
		((diffuse_coeff * O_DIFFUSE[poly][BLU_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][BLU_IND] * cos_alpha_sq)));


	/* now add other lights */
	for (il =0; il < NUM_LIGHTS;il++) {
	   /* compute phong intensity */
	   cos_theta = dot(N.p, lights[il].p);
	   cos_alpha = dot(rflxs[il].p, V.p);
	  
	   if (FP_LT(cos_alpha, 0.0)) cos_alpha_sq = 0.0;
	   else {
		cos_alpha_sq = (cos_alpha* cos_alpha) / 
		(eye_len_sq *
		(rflxs[il].p[X_COORD]*rflxs[il].p[X_COORD] + 
		 rflxs[il].p[Y_COORD]*rflxs[il].p[Y_COORD] +
		 rflxs[il].p[Z_COORD]*rflxs[il].p[Z_COORD] +
		 rflxs[il].p[W_COORD]*rflxs[il].p[W_COORD]));
           }
	   if (FP_LT(cos_theta, 0.0)) cos_theta_sq = 0.0; 
	   else {
		cos_theta_sq = (cos_theta * cos_theta) / 
		(norm_len_sq *
		(lights[il].p[X_COORD]*lights[il].p[X_COORD] + 
		 lights[il].p[Y_COORD]*lights[il].p[Y_COORD] +
		 lights[il].p[Z_COORD]*lights[il].p[Z_COORD] +
		 lights[il].p[W_COORD]*lights[il].p[W_COORD]));
           }

	   poly_color[RED_IND] = poly_color[RED_IND] + 
		(Lights[il][RED_IND] *  LIGHT_BRIGHTNESS *
		((diffuse_coeff * O_DIFFUSE[poly][RED_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][RED_IND]* cos_alpha_sq)));
	   poly_color[GRN_IND] = poly_color[GRN_IND] + 
		(Lights[il][GRN_IND] * LIGHT_BRIGHTNESS *
		((diffuse_coeff * O_DIFFUSE[poly][GRN_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][GRN_IND] * cos_alpha_sq)));
	   poly_color[BLU_IND] = poly_color[BLU_IND] + 
		(Lights[il][BLU_IND] * LIGHT_BRIGHTNESS *
		((diffuse_coeff * O_DIFFUSE[poly][BLU_IND] * cos_theta_sq) + 
		 (spec_coeff * O_SPECULAR[poly][BLU_IND] * cos_alpha_sq)));

	}

	
}

/**********************************************************/
/* FUNCTION: compute_x_deltas()			  	  */
/* PURPOSE: this function goes thru the AET table and     */
/*      computes the starting and incremental intensities */
/*      for each polygon of the current scanline 	  */
/**********************************************************/
extern void
compute_x_deltas(npols)
	GLint npols;
{
	struct ET_element *AE_trace;
	GLdouble xspan;
	GLint iz, current_poly, ic ,il;

        for (iz = 0; iz < npols;iz++)  PT[iz].inits_done = GL_FALSE;

	AE_trace = AET;
	while (AE_trace != NULL) {
		current_poly = AE_trace->poly_id;
		if (!PT[current_poly].inits_done) {
			PT[current_poly].dist_start = AE_trace->dist_start;
			
			for (ic = 0; ic < 4; ic++ ) {
			  PT[current_poly].N_start.p[ic] =
				 AE_trace->N_start.p[ic];
		   	  PT[current_poly].R_start.p[ic] =
				 AE_trace->R_start.p[ic];
		    	  PT[current_poly].V_start.p[ic] =	
				 AE_trace->V_start.p[ic];
			  PT[current_poly].L_start.p[ic] =
				 AE_trace->L_start.p[ic];
			  for (il = 0; il < NUM_LIGHTS; il++) {
			     PT[current_poly].lghts_start[il].p[ic] =
				 AE_trace->lghts_start[il].p[ic]; 
			     PT[current_poly].refls_start[il].p[ic] =
				 AE_trace->refls_start[il].p[ic];
			  }
			}
			PT[current_poly].x_start = AE_trace->x;
			PT[current_poly].inits_done = GL_TRUE;
		}
		else {
			xspan =  AE_trace->x - PT[current_poly].x_start;
			if (FP_GT(xspan, 0.0)) {

			  PT[current_poly].dist_delta = 
				(AE_trace->dist_start -
				PT[current_poly].dist_start)/xspan;
			  
			for (ic = 0; ic < 4; ic++ ) {
			  PT[current_poly].N_delta.p[ic] = 
				(AE_trace->N_start.p[ic] -
				PT[current_poly].N_start.p[ic])/xspan;
			  PT[current_poly].R_delta.p[ic] = 
				(AE_trace->R_start.p[ic] -
				PT[current_poly].R_start.p[ic])/xspan;
			  PT[current_poly].V_delta.p[ic] = 
				(AE_trace->V_start.p[ic] -
				PT[current_poly].V_start.p[ic])/xspan;
			  PT[current_poly].L_delta.p[ic] = 
				(AE_trace->L_start.p[ic] -
				PT[current_poly].L_start.p[ic])/xspan;
			  for (il = 0; il < NUM_LIGHTS; il++) {
			     PT[current_poly].lghts_delta[il].p[ic] =
				(AE_trace->lghts_start[il].p[ic] - 
				PT[current_poly].lghts_start[il].p[ic])/xspan; 
			     PT[current_poly].refls_delta[il].p[ic] =
				(AE_trace->refls_start[il].p[ic] -
				PT[current_poly].refls_start[il].p[ic])/xspan;
			  }
			 }	
			}
			else {
			  for (iz = 0; iz < 4; iz++) {
			    PT[current_poly].N_delta.p[iz] = 0.0;
		 	    PT[current_poly].R_delta.p[iz] = 0.0;
 			    PT[current_poly].V_delta.p[iz] = 0.0;
 			    PT[current_poly].L_delta.p[iz] = 0.0;
			    for (il = 0; il < NUM_LIGHTS; il++) {
				PT[current_poly].lghts_delta[il].p[iz] = 0.0;
				PT[current_poly].refls_delta[il].p[iz] = 0.0;
			    }
			  }
			  PT[current_poly].dist_delta = 0.0;
				
			}
		}
		AE_trace = AE_trace->next;
	}


#ifdef DEBUG_ET
	dfd = fopen(DEBUGFILE, "a");

	   for (iz = 0; iz < npols ; iz++) {
	    if (FP_GT(PT[iz].x_start, 0.0)) {
		fprintf(dfd, "face (%d) dist_start: %lf dist_delta: %lf\n", 
		  iz, PT[iz].dist_start, PT[iz].dist_delta);
		fprintf(dfd, "x_start: %lf\n", PT[iz].x_start);
		print_point("N_start", PT[iz].N_start);
		print_point("N_delta", PT[iz].N_delta);
		print_point("R_start", PT[iz].R_start);
		print_point("R_delta", PT[iz].R_delta); 
		print_point("V_start", PT[iz].V_start);
		print_point("V_delta", PT[iz].V_delta); 
		print_point("L_start", PT[iz].L_start);
		print_point("L_delta", PT[iz].L_delta); 
		for (il =0; il < NUM_LIGHTS; il++) {
			print_point("lghts_start", PT[iz].lghts_start[il]);
			print_point("lghts_delta", PT[iz].lghts_delta[il]);
			print_point("refls_start", PT[iz].refls_start[il]);
			print_point("refls_delta", PT[iz].refls_delta[il]);
		}
	    }
	}
#endif
	
}
 
/**********************************************************/
/* FUNCTION: dot()                                        */
/* PURPOSE: this function computes the dot product of 2   */
/*              4 D vectors                               */
/**********************************************************/
extern GLdouble
dot(a, b)
        GLdouble a[4], b[4];
{        
        GLdouble  return_val;
 
        return_val = a[X_COORD]*b[X_COORD] + a[Y_COORD]*b[Y_COORD] +
                     a[Z_COORD]*b[Z_COORD] + a[W_COORD]*b[W_COORD];

        return (return_val);                
}

/**********************************************************/
/* FUNCTION: depth_q()              	 		   */
/* PURPOSE: this function computes the transparency       */
/*      factors by distance (clearnest = closeness)       */
/*              It uses the atmospheric attenuation       */
/*              equations on pp. 727-728 of Foley and     */
/*              van Dam                                   */
/**********************************************************/
extern void
depth_q()
{
        GLint ret_val, iface;
      
         z_back = polytope->vertex[polytope->ordered_vertices[0]].depth;
         z_front = polytope->vertex[
	 	polytope->ordered_vertices[polytope->num_vertices-1]].depth;
 
	if (FP_EQ(z_front, z_back))  z_back = z_front + .1;
          depth_range = z_back - z_front;
          z_front = z_back - Z_FRONT_FRACTION*depth_range;
          z_back = z_back + Z_BACK_FRACTION*depth_range;

#ifdef DEBUG_SCAN
	dfd = fopen(DEBUGFILE, "a");
	fprintf(dfd, "z_front: %lf z_back: %lf\n", z_front, z_back);
	fclose(dfd);
#endif
		  
}

