/*
 * FILE: gx.c
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this program  handles all the graphics library calls 
 */

#include <stdio.h> 
#include <math.h>
#include <GL/aux.h> 
#include "view4d.h"

#define INIT_X		0
#define INIT_Y		0

GLint fd;
/*********** function declarations *********************/
static void initialize();
static void Display_Func();

void
main(argc, argv)
	GLint argc;
	char *argv[];
{
	FILE *fp;
	struct rle_header_struct header;


 	if (argc == 2) {
		fp = fopen(argv[1], "r");

		fd = fileno(fp);

        	if (read(fd, (char *) &header, sizeof(struct rle_header_struct)) !=
                        sizeof(struct rle_header_struct))
        	{
                	perror("read");
                	exit(1);
        	}
		
		initialize(header.width, header.height);
	}
   	 else {
               printf("Usage: %s <description-file> !!!!\n", argv[0]);
               exit(1);
                
    	}  

	sleep(60000);     
 
}


static void
Display_Func()
{
        GLint y, r, g, b; 
	GLfloat parray[MAX_PIXELS][3]; 
	GLint len, i, bytes, tot_len;
	GLuint pixel;
	struct rle_header_struct header;

	lseek(fd, 0, SEEK_SET);

	if (read(fd, (char *) &header, sizeof(struct rle_header_struct)) != 
			sizeof(struct rle_header_struct))
	{
		perror("read");
		exit(1);
	}

	for (y = 0;y < header.height;y++) {

	     tot_len = 0;
 	     do
             {  
	        bytes = read(fd, (char *) &pixel, 4);
		len = pixel>>24; 
		pixel = pixel - (len<<24); 
		for (i=0;i<=len;i++) {
			b = pixel>>16;
			g = (pixel - (b<<16))>>8;
			r = pixel - (b<<16) - (g<<8);
			parray[i][0] = r/255.;
			parray[i][1] = g/255.;
			parray[i][2] = b/255.;
		}
		glRasterPos2i(tot_len, y);
		glDrawPixels(len + 1, 1, GL_RGB, GL_FLOAT, parray[0]);
		tot_len = tot_len + len + 1;
	     } while (tot_len < header.width);
	

	     if (!(y % 50)) glFlush();
	}
	glFlush();
}


static GLenum Key(int key, GLenum mask)
{

    switch (key) {
      case AUX_ESCAPE:
      case AUX_q:
      case AUX_Q:
                    auxQuit();
                    break;
      default:
		break;
    }
    return GL_TRUE;
}

/**********************************************************/
/* FUNCTION: initialize()                                 */
/* PURPOSE:   this sets up all the graphics settings and  */
/*      devices.                                          */
/**********************************************************/
static void
initialize(GLint width, GLint height)
{
        GLenum type;
        GLint x, y;

        auxInitPosition(INIT_X, INIT_Y, width, height);

        type = AUX_RGB | AUX_SINGLE | AUX_INDIRECT;
        auxInitDisplayMode(type);


        if (auxInitWindow("4D Polytope Previewer") == GL_FALSE) {
                printf("cannot InitWindow\n");
                auxQuit();
        }
        glShadeModel(GL_FLAT);
        glDisable(GL_DEPTH_TEST);
	glClearColor(0., 0., 0., 0.);
	glDrawBuffer(GL_FRONT);
        glClear(GL_COLOR_BUFFER_BIT);

	gluOrtho2D(0, width, 0, height);

        auxMainLoop(Display_Func);

}
