/*
 * FILE: scan.c     
 * PROGRAMMER: Tom Asbury 
 * PURPOSE: this program scans converts a given set of edges
*/

#include <stdio.h> 
#include <math.h>
#include <malloc.h>
#include <GL/aux.h> 
#include "view4d.h"

/******************** define light source positions *************/
#define LIGHT_X 			vx
#define LIGHT_Y				vy
#define LIGHT_Z				vz
#define LIGHT_W				vw
GLdouble lights[][4] = {
	{50., 0., 50., 0. },
	{-35., 35., 50., 0. },
	{-35., -35., 50., 0. },
	{-50., 0., -50., 0. },
	{35., 35., -50., 0. },
	{35., -35., -50., 0. },
	{0., 0., 0., 50. },
};

/********** external variable definitions ****************/
struct PT_struct *PT;
struct ET_element **ET;
struct ET_element *AET;
	
/*********************************************************/

/*********** function declarations *********************/
extern void initialize_edge_table();
static void add_edge_to_ET();
extern void add_face_to_ET();
extern void sort_edge_table();
static struct ET_element *sort_ll();
static struct ET_element *merge_ll();
extern void do_scan_convert();
extern void fill_pixels();
static void AET_edge_removal();
extern void draw_scanline();
static void update_AET();
extern void print_ET();
static void print_ET_element();
extern void print_point();
extern void normalize();
extern void compute_x_deltas();
extern GLdouble dot();
extern GLint cross();
/********************************************************/

void
initialize_edge_table()
/********************************************************/
/* FUNCTION: intialize_edge_table			*/
/* PURPOSE:  sets all edges in edge table to NULL	*/
/********************************************************/
{
	GLint iscan;
	struct ET_element *ET_trace, *next_trace;

	for (iscan = 0;iscan <= height_of_window;iscan++ ) {
		ET_trace = ET[iscan];
		ET[iscan] = NULL;
		while (ET_trace != NULL) {
			next_trace = ET_trace->next;
			ET_trace = NULL;
			ET_trace = next_trace;
		}
	}
	ET_trace = AET;
	AET = NULL;
	while (ET_trace != NULL) {
			next_trace = ET_trace->next;
			ET_trace = NULL;
			ET_trace = next_trace;
	}
}

void
add_face_to_ET(iface, ipoly)
	GLint iface, ipoly;
/********************************************************/
/* FUNCTION: add_face_to_ET				*/
/* PURPOSE:  this computes adds a given polytope face's  */
/*  edge to the edge_table 				*/
/********************************************************/
{
	GLint iedge, edge_num;
	struct edge_struct edge;

	for (iedge = 0;iedge<polytope->face[iface].num_edges; iedge++) 
	{
	    edge_num = polytope->face[iface].edge[iedge];
	    edge = polytope->edge[edge_num];
	    add_edge_to_ET(iface, ipoly, edge);

	}	
}

static void
add_edge_to_ET(iface, ipoly, edge)
	GLint iface, ipoly;
	struct edge_struct edge;
/********************************************************/
/* FUNCTION: add_edge_to_ET				*/
/* PURPOSE:  appends a given edge to the edge table     */
/********************************************************/
{
	struct point a, b, L_start, L_end, LLL, N_end, R_end, V_end;
	struct point *pa, *pb;
	struct ET_element *ET_trace, *new_ET_element;
	GLint num_vtx, scanline, ic, il, high, low;
	GLdouble ymin, ymax, dst, scan_span, dot_product, dist_end;

	num_vtx = edge.a;
	for (ic = 0; ic < 4; ic++ ) 
	   a.p[ic] = polytope->screen_vtx[num_vtx].p[ic];
	num_vtx = edge.b;
	for (ic = 0; ic < 4; ic++ ) 
	   b.p[ic] = polytope->screen_vtx[num_vtx].p[ic];

	if ((int) ceil(a.p[Y_COORD]) != (int) ceil(b.p[Y_COORD])) {
		/* edge is NOT horizontal */
		ymin = min(a.p[Y_COORD],b.p[Y_COORD]);
		scanline = (int) ceil(ymin);
		ET_trace = ET[scanline];
		
		if (ET_trace != NULL) {
			new_ET_element = (struct ET_element*) 
				malloc(sizeof(struct ET_element));
			/* find end of list */
			while (ET_trace->next != NULL) 
			   ET_trace = ET_trace->next;
			ET_trace->next = new_ET_element;
			ET_trace = ET_trace->next;
	        }
		else {
			/* add first edge to list */
			ET_trace = (struct ET_element *)
				malloc(sizeof(struct ET_element));
			ET[scanline] = ET_trace;
		}
						
	        if (FP_EQ(ymin, a.p[Y_COORD])) {
			/* a is lower endpoint */
			ET_trace->x = a.p[X_COORD];
			ymax = b.p[Y_COORD];
			low = edge.a;
			high = edge.b;
		}
		else {
			/* b is lower endpoint */
			ET_trace->x = b.p[X_COORD];
			ymax = a.p[Y_COORD];
			low = edge.b;
			high = edge.a;
		}
		ET_trace->ymax = ymax ;
		if (fabs(a.p[X_COORD] - b.p[X_COORD]) <= EPSILON)
			ET_trace->x_incr = 0.0;
		else 
			ET_trace->x_incr = (double) (1./slope(a,b));
		ET_trace->poly_id = ipoly;
		ET_trace->face_id = iface;
		ET_trace->edge.a = edge.a;
		ET_trace->edge.b = edge.b;
		ET_trace->next = NULL;
		ET_trace->dist_start =  DIST(
		  polytope->rot_vertex[low].p[X_COORD], 
		   polytope->centroid.p[X_COORD], 
		  polytope->rot_vertex[low].p[Y_COORD],
		   polytope->centroid.p[Y_COORD], 
		  polytope->rot_vertex[low].p[Z_COORD],
		   polytope->centroid.p[Z_COORD],
		  polytope->rot_vertex[low].p[W_COORD],
		   polytope->centroid.p[W_COORD]);
 		dist_end =  DIST(
		  polytope->rot_vertex[high].p[X_COORD], 
		    polytope->centroid.p[X_COORD], 
		  polytope->rot_vertex[high].p[Y_COORD], 
		    polytope->centroid.p[Y_COORD], 
		  polytope->rot_vertex[high].p[Z_COORD], 
		    polytope->centroid.p[Z_COORD],
		  polytope->rot_vertex[high].p[W_COORD], 
		    polytope->centroid.p[W_COORD]);


		/*  place N, R, L and V in ET_struct */

		/* do start of edge first (low) */
	        for (ic = 0; ic < 4; ic++ ) {
	           if (cross()) ET_trace->N_start.p[ic] = 
			PT[ipoly].gouraud_norm.p[ic];
		   else ET_trace->N_start.p[ic] =
			polytope->gouraud_norm[low].p[ic];
		   a.p[ic] = polytope->rot_vertex[low].p[ic];
		 }
 
		L_start.p[X_COORD] = LIGHT_X - a.p[X_COORD];
		L_start.p[Y_COORD] = LIGHT_Y - a.p[Y_COORD];
		L_start.p[Z_COORD] = LIGHT_Z - a.p[Z_COORD];
		L_start.p[W_COORD] = LIGHT_W - a.p[W_COORD];
	        ET_trace->V_start.p[X_COORD] = vx - a.p[X_COORD];
		ET_trace->V_start.p[Y_COORD] = vy - a.p[Y_COORD];
		ET_trace->V_start.p[Z_COORD] = vz - a.p[Z_COORD];
		ET_trace->V_start.p[W_COORD] = vw - a.p[W_COORD];
		normalize(L_start.p);
		normalize(ET_trace->V_start.p);
		for (ic = 0; ic < 4; ic++) 
		   ET_trace->L_start.p[ic] = L_start.p[ic];
		dot_product = dot(ET_trace->N_start.p, L_start.p);
		for (ic = 0; ic < 4; ic++) 
			ET_trace->R_start.p[ic] = (2.*ET_trace->N_start.p[ic] 
			   * dot_product) - L_start.p[ic];
		normalize(ET_trace->R_start.p);
		/* now do other light sources */
		for (il = 0; il < NUM_LIGHTS; il++) {
		  for (ic = 0; ic < 4; ic++)
		        LLL.p[ic] = lights[il][ic] - a.p[ic];
		  normalize(LLL.p);
		  for (ic = 0; ic < 4; ic++)
			ET_trace->lghts_start[il].p[ic] = LLL.p[ic];
		  dot_product = dot(ET_trace->N_start.p, LLL.p);
		  for (ic = 0; ic < 4; ic++)
			ET_trace->refls_start[il].p[ic] = (2. * 
			ET_trace->N_start.p[ic] *
			dot_product) - LLL.p[ic];
		  normalize(ET_trace->refls_start[il].p);
	       }
 

		/* now do end of edge (high) */
		for (ic = 0; ic < 4; ic++) {
	           if (cross()) N_end.p[ic] = 
			PT[ipoly].gouraud_norm.p[ic];
		   else	  N_end.p[ic] = 
			polytope->gouraud_norm[high].p[ic];
		   b.p[ic] = polytope->rot_vertex[high].p[ic];
		}
		L_end.p[X_COORD] = LIGHT_X - b.p[X_COORD];
		L_end.p[Y_COORD] = LIGHT_Y - b.p[Y_COORD];
		L_end.p[Z_COORD] = LIGHT_Z - b.p[Z_COORD];
		L_end.p[W_COORD] = LIGHT_W - b.p[W_COORD];
		V_end.p[X_COORD]= vx - b.p[X_COORD];
		V_end.p[Y_COORD]= vy - b.p[Y_COORD];
		V_end.p[Z_COORD]= vz - b.p[Z_COORD];
	        V_end.p[W_COORD]= vw - b.p[W_COORD];
		normalize(L_end.p);
		normalize(V_end.p);
		dot_product = dot(N_end.p, L_end.p);
		for (ic = 0; ic < 4; ic++) 
		  R_end.p[ic] = (2. * N_end.p[ic] * dot_product) - L_end.p[ic];
		normalize(R_end.p);
		/* now do other light sources */
		for (il = 0; il < NUM_LIGHTS; il++) {
		  for (ic = 0; ic < 4; ic++) LLL.p[ic] = 
				lights[il][ic] - b.p[ic];
		  normalize(LLL.p);
		  for (ic = 0; ic < 4; ic++)
		     ET_trace->lghts_incr[il].p[ic] = LLL.p[ic];
		  dot_product = dot(N_end.p, LLL.p);
		  for (ic = 0; ic < 4; ic++)
                    ET_trace->refls_incr[il].p[ic] = (2. * N_end.p[ic] *
			  dot_product) - LLL.p[ic];
		  normalize(ET_trace->refls_incr[il].p);
	       }

#ifdef DEBUG_ET
		fprintf(dfd," EDGE %d -----> %d\n", low, high);
		fprintf(dfd," \t N (%lf,%lf,%lf,%lf) ---->"
			" \n\t  (%lf,%lf,%lf,%lf)\n",
		   ET_trace->N_start.p[X_COORD], ET_trace->N_start.p[Y_COORD],
		   ET_trace->N_start.p[Z_COORD], ET_trace->N_start.p[W_COORD],
			N_end.p[X_COORD], N_end.p[Y_COORD],
			N_end.p[Z_COORD], N_end.p[W_COORD]);
		fprintf(dfd," \t R (%lf,%lf,%lf,%lf) ---->"
			" \n\t  (%lf,%lf,%lf,%lf)\n",
		   ET_trace->R_start.p[X_COORD], ET_trace->R_start.p[Y_COORD],
		   ET_trace->R_start.p[Z_COORD], ET_trace->R_start.p[W_COORD],
			R_end.p[X_COORD], R_end.p[Y_COORD],
			R_end.p[Z_COORD], R_end.p[W_COORD]);
		fprintf(dfd," \t L (%lf,%lf,%lf,%lf) ---->"
			" \n\t   (%lf,%lf,%lf,%lf)\n",
		   ET_trace->L_start.p[X_COORD], ET_trace->L_start.p[Y_COORD],
		   ET_trace->L_start.p[Z_COORD], ET_trace->L_start.p[W_COORD],
			L_end.p[X_COORD], L_end.p[Y_COORD],
			L_end.p[Z_COORD], L_end.p[W_COORD]);
		fprintf(dfd," \t V (%lf,%lf,%lf,%lf) ---->"
			" \n\t   (%lf,%lf,%lf,%lf)\n",
		   ET_trace->V_start.p[X_COORD], ET_trace->V_start.p[Y_COORD],
		   ET_trace->V_start.p[Z_COORD], ET_trace->V_start.p[W_COORD],
			V_end.p[X_COORD], V_end.p[Y_COORD],
			V_end.p[Z_COORD], V_end.p[W_COORD]);
		fprintf(dfd, " N_start * L_start = %lf\n",
			dot(ET_trace->N_start.p, ET_trace->L_start.p));
		fprintf(dfd, " N_end * L_end = %lf\n",
			dot(N_end.p, L_end.p));
		fprintf(dfd, " R_start * V_start = %lf\n",
			dot(ET_trace->R_start.p, ET_trace->V_start.p));
		fprintf(dfd, " R_end * V_end = %lf\n",
			dot(R_end.p, V_end.p));
		for (il = 0; il < NUM_LIGHTS; il++) {
			fprintf(dfd," \t L[%d] (%lf,%lf,%lf,%lf) ---->"
			  " \n\t   (%lf,%lf,%lf,%lf)\n",
		   	   il, 
			  ET_trace->lghts_start[il].p[X_COORD],
			  ET_trace->lghts_start[il].p[Y_COORD],
			  ET_trace->lghts_start[il].p[Z_COORD], 
			  ET_trace->lghts_start[il].p[W_COORD],
			  ET_trace->lghts_incr[il].p[X_COORD], 
			  ET_trace->lghts_incr[il].p[Y_COORD],
			  ET_trace->lghts_incr[il].p[Z_COORD], 
			  ET_trace->lghts_incr[il].p[W_COORD]);
			fprintf(dfd," \t R[%d] (%lf,%lf,%lf,%lf) ---->"
			  " \n\t   (%lf,%lf,%lf,%lf)\n",
		   	  il, 
			  ET_trace->refls_start[il].p[X_COORD],
			  ET_trace->refls_start[il].p[Y_COORD],
			  ET_trace->refls_start[il].p[Z_COORD], 
			  ET_trace->refls_start[il].p[W_COORD],
			  ET_trace->refls_incr[il].p[X_COORD], 
			  ET_trace->refls_incr[il].p[Y_COORD],
			  ET_trace->refls_incr[il].p[Z_COORD], 
			  ET_trace->refls_incr[il].p[W_COORD]);
		}
#endif

		scan_span = ceil(ymax)  - ceil(ymin);
		ET_trace->dist_incr = (dist_end -
			ET_trace->dist_start)/scan_span;
		
		for (ic = 0; ic < 4; ic++ ) {
		  ET_trace->N_incr.p[ic] = (N_end.p[ic] - 
		    ET_trace->N_start.p[ic])/scan_span;
		  ET_trace->R_incr.p[ic] = (R_end.p[ic] - 
		    ET_trace->R_start.p[ic])/scan_span;
		  ET_trace->V_incr.p[ic] = (V_end.p[ic] - 
		    ET_trace->V_start.p[ic])/scan_span;
		  ET_trace->L_incr.p[ic] = (L_end.p[ic] - 
		    ET_trace->L_start.p[ic])/scan_span;
		  /* now do other light sources */
		  for (il = 0; il < NUM_LIGHTS; il++ ) {
		    ET_trace->lghts_incr[il].p[ic] = 
			(ET_trace->lghts_incr[il].p[ic] - 
		    	ET_trace->lghts_start[il].p[ic])/scan_span;
  		    ET_trace->refls_incr[il].p[ic] = 
			(ET_trace->refls_incr[il].p[ic] - 
		    	ET_trace->refls_start[il].p[ic])/scan_span;
		  }
		}
 		  		   
	}
}
		
extern void
sort_edge_table()
/********************************************************/
/* FUNCTION: sort_edge_table				*/
/* PURPOSE:  this sorts each scan line of the edge table*/
/*  by the x value of each edge by using merge sort	*/
/********************************************************/
{
	struct ET_element *trace;
	GLint i, iscan, length;

	for (iscan = 0; iscan <= height_of_window;iscan++) {
		if (ET[iscan] != NULL) {
		  /* first find number of elements in linked list */
		  trace = ET[iscan];
		  length = 0;
		  while (trace != NULL) {
			length++;
			trace = trace->next;
		  }
		  /* now sort these elements by x */
		  ET[iscan] = sort_ll(ET[iscan],length);
		}
	}
}

/********************************************************/
/* FUNCTION: sort_ll()  				*/
/* PURPOSE:  this function sorts a linked list of edges */
/*	      by spliting the list in half and calling  */
/*	      merge sort. It is recursive		*/
/********************************************************/

static struct ET_element
*sort_ll(head, length)
	struct ET_element *head;
	GLint length;
{
	struct ET_element *a, *b, *c;
	GLint halfway, i;

	if (head == NULL) return(head);
	if (length == 1) return(head);
	a = head;b = head;
	halfway = length/2;
	for (i = 2; i <= halfway; i++) head = head->next;
	b = head->next;head->next = NULL;
	return(merge_ll(sort_ll(a, halfway), sort_ll(b, length - halfway)));
}

/********************************************************/
/* FUNCTION: merge_ll					*/
/* PURPOSE:  this function merges two sorted linked     */
/*           first by the "x" value and then using      */
/*	     "1/m" value to break ties.			*/
/********************************************************/

static struct ET_element
*merge_ll(a, b)
	struct ET_element *a, *b;
{
	struct ET_element *c, *head;
	GLboolean done;

	if ((a == NULL) && (b == NULL)) return(a);
	/* point c to head of list */
	   if ((a != NULL) && (b == NULL)) return(a);
	   if ((b != NULL) && (a == NULL)) return(b);
           if ((b->x - a->x) > EPSILON)  {
		c = a;a = a->next;       }   
	   else if ((a->x - b->x) > EPSILON) {
		c = b;b = b->next;       }
	   else if ((b->x_incr - a->x_incr) > EPSILON)  {
		c = a;a = a->next;			}
	   else {  c = b;b = b->next; }
	head = c;
	done = GL_FALSE;
	while (!done) {
	  if ((a == NULL) && (b != NULL)) {
	    if (b == NULL) done = GL_TRUE;
	    else {
	     /* copy the rest of b to merged list */
	     c->next = b;c = b; b = b->next;
	     while (b != NULL) {
		c->next = b;c = b; b = b->next;
	     }
	     done = GL_TRUE;
	    }
	  }
	  else if ((b == NULL) && (a != NULL)) {
	    /* copy the rest of a to merged list */
	    c->next = a;c = a; a = a->next;
	    while (a != NULL) {
		c->next = a;c = a; a = a->next;
	    }
	    done = GL_TRUE;
	  }
	  else if ((b->x - a->x) > EPSILON) {
		c->next = a; c = a; a = a->next;
	  } 
	  else if ((a->x - b->x) > EPSILON) {
		c->next = b; c = b; b = b->next;
	  }
	  else {
		/* same x value so break tie on y */
		if ((b->x_incr - a->x_incr) > EPSILON) {
                	c->next = a; c = a; a = a->next;
          	} 
		else {
                	c->next = b; c = b; b = b->next;
		}
          }

	}

	return (head);
}


extern void
do_scan_convert(npols)
	GLint npols;
/********************************************************/
/* FUNCTION: do_scan_convert				*/
/* PURPOSE:  this function does the actual scan         */
/*    conversion of a given edge table using the        */
/*    algorithm given in Foley & van Dam on pp.		*/
/*	680 - 686.					*/
/********************************************************/
{
	GLint scanline, i, iset, AET_lngth;
	struct ET_element *AET_trace;

	/* set y to lowest entry in edge table */
	scanline = 0;
	while ((ET[scanline] == NULL) && (scanline < t)) 
		scanline++;
	AET = NULL;
	for (i = scanline; i < height_of_window; i++) {

		/* set all polygons to OUT */
		for (iset = 0; iset<npols; iset++) PT[iset].IN = GL_FALSE;
		(void) AET_edge_removal(i);
		(void) update_AET(i);
		AET = merge_ll(AET, ET[i]);
		AET_lngth = 0;
		AET_trace = AET;
		while (AET_trace != NULL) {
			AET_lngth++;
			AET_trace = AET_trace->next;
		}
		AET = sort_ll(AET, AET_lngth);
		if ((AET != NULL) && (i >= 0))
			 draw_scanline(i, npols);
		if ((i % 10) == 0) glFlush();

	}
}

static void
AET_edge_removal(Y)
	GLint Y;
/********************************************************/
/* FUNCTION: AET_edge_removal				*/
/* PURPOSE:  this function removes those edges from     */
/*       the AET that have a y_max <= to the current    */
/*       scanline "Y".				        */ 
/********************************************************/
{
	struct ET_element *AET_trace, *previous_trace;

	AET_trace = AET;
	previous_trace = NULL;
	while (AET_trace != NULL) {
		if ((Y - AET_trace->ymax) >= -1.*EPSILON) {
			if (previous_trace == NULL) AET = AET_trace->next;
			else previous_trace->next = AET_trace->next;
		}	
		else previous_trace = AET_trace;
		AET_trace = AET_trace->next;
	}
}

static void
update_AET(i)
	GLint i;
/********************************************************/
/* FUNCTION: update_AET					*/
/* PURPOSE:  this function update all edges on the AET  */
/*	by incrementing x by x_incr			*/
/********************************************************/
{	
	struct ET_element *AET_trace;
	GLint ic, il;

	AET_trace = AET;
	while (AET_trace != NULL) {
		AET_trace->x = AET_trace->x + AET_trace->x_incr;
		AET_trace->dist_start = AET_trace->dist_start +
		     AET_trace->dist_incr;

		for (ic = 0; ic < 4; ic++) {
		  AET_trace->N_start.p[ic] = 
      		     AET_trace->N_start.p[ic] +
		     AET_trace->N_incr.p[ic];
		  AET_trace->R_start.p[ic] = 
      		     AET_trace->R_start.p[ic] +
		     AET_trace->R_incr.p[ic];
		  AET_trace->V_start.p[ic] = 
      		     AET_trace->V_start.p[ic] +
		     AET_trace->V_incr.p[ic];
		  AET_trace->L_start.p[ic] = 
      		     AET_trace->L_start.p[ic] +
		     AET_trace->L_incr.p[ic];
		  /* now do other lights */
		  for (il =0; il < NUM_LIGHTS; il++) {
			AET_trace->lghts_start[il].p[ic] = 
  			  AET_trace->lghts_start[il].p[ic] +
		          AET_trace->lghts_incr[il].p[ic];
			AET_trace->refls_start[il].p[ic] = 
  			  AET_trace->refls_start[il].p[ic] +
		          AET_trace->refls_incr[il].p[ic];
		  }
		}

		AET_trace = AET_trace->next;
		
	}

}

static void
draw_scanline(Y, npols)
	GLint Y, npols;
/********************************************************/
/* FUNCTION: do_scan_convert				*/
/* PURPOSE:  this function does the actual drawing      */
/*    of a single scanline "Y" by looking at AET[Y]	*/
/********************************************************/
{
	GLint i, min_poly, all_false, current_poly = -1;
	GLdouble z_dist, min_dist, up_to_x;
	struct ET_element *AET_trace, *AET_E1, *AET_E2;
	
 	(void) compute_x_deltas(npols); 

	AET_E1 = AET;AET_E2 = AET_E1->next;
		 
	current_poly = AET_E1->poly_id;

#ifdef DEBUG_PIXEL
	dfd = fopen(DEBUGFILE, "a");
#endif	

	PT[current_poly].IN = GL_TRUE;
	while (AET_E2 != NULL) {
	   if (current_poly != -1) {
		/* inside a polygon */
		fill_pixels(Y, AET_E1->x, AET_E2->x);
		up_to_x = AET_E2->x;
		AET_E1= AET_E2;
		AET_E2 = AET_E1->next;
		if (AET_E2 != NULL) {
		  if (AET_E1->poly_id == current_poly) {
			/* exiting current polygon */
			PT[current_poly].IN = GL_FALSE;
		   	all_false = GL_TRUE;
			i = 0; 
		  	while ((i < npols) && all_false) {
			  if (PT[i].IN == GL_TRUE) {
				current_poly = i;
				all_false = GL_FALSE;
			  }	
			  i++;
			}
		  	if (all_false) current_poly = -1;
		        
		  }
		  else 
			/* not exiting current polygon */
			PT[AET_E1->poly_id].IN = !(PT[AET_E1->poly_id].IN);
		}
	  }
	  else {
	      /* outside of all polygons - so go into next one*/
		AET_E1 = AET_E2;
		AET_E2 = AET_E1->next;
		current_poly = AET_E1->poly_id;
		PT[current_poly].IN = GL_TRUE;
	  }
	}

#ifdef DEBUG_PIXEL
	fclose(dfd);
#endif

#ifdef DEBUG_ET
	dfd = fopen(DEBUGFILE, "a");
       fprintf(dfd, "AET[%d]:\n",Y);
	   AET_trace = AET;
	   while (AET_trace != NULL) {
		print_ET_element(AET_trace);
		AET_trace = AET_trace->next;  } 
	fclose(dfd);
#endif
}

extern void
print_ET()
{
	struct ET_element *trace;
	GLint i;

	fprintf(dfd, " * * * EDGE TABLE * * * \n\n");
	/* print out ET for now */
	for (i = b; i <= t; i++) {
		if (ET[i] != NULL) {
			trace = ET[i];
			while (trace != NULL) {
				fprintf(dfd, "ET[%d]:\n",i);
				print_ET_element(trace);
				trace = trace->next;
			}
		}
	}
}

static void
print_ET_element(a)
	struct ET_element *a;
{
	fprintf(dfd,"   edge: %d %d\n",a->edge.a, a->edge.b);
	fprintf(dfd,"    \t  x: %lf\n",a->x);
	fprintf(dfd,"   \t ymax: %lf\n",a->ymax);
	fprintf(dfd," \t x_incr: %lf\n",a->x_incr);
	fprintf(dfd,"  \t  face: %d\n",a->face_id);
	fprintf(dfd,"  \t  poly: %d\n",a->poly_id);
	   print_point("N_start",a->N_start);
	   print_point("N_incr",a->N_incr);
	   print_point("R_start",a->R_start);
	   print_point("R_incr",a->R_incr);
	   print_point("V_start",a->V_start);
	   print_point("V_incr",a->V_incr);
	   print_point("L_start",a->L_start);
	   print_point("L_incr",a->L_incr);

}

