/* FILE: view.h
 * PROGRAMMER: Tom Asbury
 * PURPOSE: this file contains external structures used in view.c
 */

#ifdef DEBUG_ALL
#define 	DEBUG_SCAN	1
#define 	DEBUG_PIXEL	1
#define 	DEBUG_ET	1
#define 	DEBUG_CLIP	1
#endif

#define X_COORD			0
#define X_COMP 			0
#define Y_COORD			1
#define Y_COMP			1
#define Z_COORD			2
#define Z_COMP			2
#define W_COORD			3
#define W_COMP			3
#define RED_IND			0
#define GRN_IND			1
#define BLU_IND			2
#define NUM_SCANLINES           height_of_window + 1
#define NUM_LIGHTS		7
#define MAX_LIGHTS              9
#define MAX_PIXELS		2048	/* max size of images */
#define MAX_FACES		1200
#define MAX_VERTICES		10000
#define MAX_EDGES		MAX_VERTICES
#define MAX_FACE_VERTICES	20
#define MAX_FACE_EDGES		MAX_FACE_VERTICES
#define MAX_FILE_NAME_LEN	100

#define EPSILON       3.0517578e-11 /* error on double point ops */
#define DIST(x1,x2,y1,y2,z1,z2,w1,w2)  \
		sqrt(pow(x1-x2,2.)+pow(y1-y2,2.)+pow(z1-z2,2.)+pow(w1-w2,2.))
#define FP_EQ(x,y)        ((((x) - (y))<=EPSILON)&&(((x) - (y))>=-1.*EPSILON))
#define FP_NEQ(x,y)        ((((x) - (y))>EPSILON)||(((x) - (y))<-1.*EPSILON))
#define PXFP_EQ(p1, p2)    FP_EQ(p1.p[X_COORD],p2.p[X_COORD])
#define PYFP_EQ(p1, p2)    FP_EQ(p1.p[Y_COORD],p2.p[Y_COORD])
#define PZFP_EQ(p1, p2)    FP_EQ(p1.p[Z_COORD],p2.p[Z_COORD])
#define PWFP_EQ(p1, p2)    FP_EQ(p1.p[W_COORD],p2.p[W_COORD])
#define FP_GT(x,y)        ((x) - (y) > EPSILON)
#define FP_LT(x,y)        ((y) - (x) > EPSILON)
#define INF		9.9999999e11
#define DEBUGFILE 		"outpoly"

#define min(x, y) (x)>(y) ? (y) : (x)
#define slope(a,b) ((b.p[Y_COORD]-a.p[Y_COORD])/(b.p[X_COORD]-a.p[X_COORD]))

struct rle_header_struct {
        int     width;
        int     height;
        int     bits_per_pixel;
};

struct point {
	GLdouble p[4];
	GLdouble depth;
};

struct edge_struct {
	GLint a;
	GLint b;
};

struct face_struct {
	GLint num_vertices;
	GLint *vertex;
	GLint num_edges;
	GLint *edge;
	GLint visible;
	GLdouble depth;
	struct point surface_normal;
};

struct cell_struct {
	GLint num_faces;
	struct face_struct *face;
	struct point volume_normal;
	GLdouble depth;
};

/************************************************************************/

    	/**********************************************************/
	/* the following is the polytope structure used to house   */
	/* all the information necessary to graph the polytope     */
	/* from all perspectives. 				  */
	/**********************************************************/

struct polytope_struct {
	struct edge_struct *edge;     /* pointer to edge table */
	GLint num_edges;
        struct point *vertex; 	      /* pointer to vertex table */
 	struct point *rot_vertex;     /* pointer to vertex table */
	struct point *v3D;	      /* pointer to projected vertices */
	struct point *screen_vtx;     /* pointer to vtx screen coords */
	struct point *gouraud_norm;   /* pointer to vtx screen coords */		
	struct point centroid;      /* pointer to centroid of polytope */
	GLint num_vertices;
        struct face_struct *face;     /* pointer to face table */
	GLint num_faces;
	struct cell_struct *cell;
	GLint num_cells;
	GLboolean	clipped;
	struct point origin;	       /* x, y, z, w position of of polytope */
	GLdouble scale_fac;	      /* scaling factor for size of polytope */
	GLint *ordered_faces;	      /* listing of faces from nearest to far */
	GLint *ordered_vertices;       /* listing of vtxs from nearest to far */
};
 
        /**********************************************************/
        /* the following is the polygon_table structure used to   */
        /* keep all the information necessary to determine visible*/
        /* surface removal for all the polygons                   */
        /**********************************************************/

struct PT_struct {
        struct point N_start;
        struct point N_delta;
        struct point R_start;
        struct point R_delta;
        struct point refls_start[MAX_LIGHTS];
        struct point refls_delta[MAX_LIGHTS];
	struct point V_start;
        struct point V_delta;
        struct point L_start;
        struct point L_delta;
   	struct point lghts_start[MAX_LIGHTS];
        struct point lghts_delta[MAX_LIGHTS];
	struct point gouraud_norm;
        GLdouble dist_start;
        GLdouble dist_delta;
        GLdouble x_start;
        GLint IN;
	GLint inits_done;
	GLdouble transp_fac;
	GLdouble intensity;
};
       /**********************************************************/
        /* the following structre is used to construct the edge   */
        /* table which is  necessary to scan convert a list of    */
        /* given vertices                                         */
        /**********************************************************/
 
struct ET_element {
        struct edge_struct edge;
        GLdouble x;
        GLdouble z;
        GLdouble ymax;
        GLdouble x_incr;
        struct point N_start;
        struct point N_incr;
        struct point R_start;
        struct point R_incr;
        struct point refls_start[MAX_LIGHTS];
        struct point refls_incr[MAX_LIGHTS];
	struct point V_start;
        struct point V_incr;
        struct point L_start;
        struct point L_incr;
        struct point lghts_start[MAX_LIGHTS];
        struct point lghts_incr[MAX_LIGHTS];
	double dist_start;
        double dist_incr;
        GLint poly_id;
        GLint obj_id;
        GLint face_id;
        struct ET_element *next;
};
	
/********************************************************************/

extern struct PT_struct *PT;
extern struct ET_element **ET;
extern struct ET_element *AET;
extern GLint r, l, t, b, n, f;
extern GLdouble persp_mat[16];
extern struct polytope_struct *polytope;
extern GLboolean archive_it; 
extern FILE *dfd;
extern GLdouble vx, vy, vz, vw, close_tsp, far_tsp, close_amb, far_amb;
extern GLdouble close_diffuse, far_diffuse, close_spec, far_spec;
extern GLdouble O_DIFFUSE[][3], O_SPECULAR[][3];
extern GLint  image[MAX_PIXELS][MAX_PIXELS];
extern char input_file[], archive_name[];
extern GLint width_of_window, height_of_window;
